@@ -1,5 +1,5 @@
 # Relational
-# Copyright (C) 2008-2016  Salvo "LtWorf" Tomaselli
+# Copyright (C) 2008-2020 Salvo "LtWorf" Tomaselli
 #
 # Relational is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -24,11 +24,36 @@
 #
 # Language definition here:
 # http://ltworf.github.io/relational/grammar.html
+from typing import Optional, Union, List, Any, Dict, Literal
+from dataclasses import dataclass
+
 from relational import rtypes
 
-RELATION = 0
-UNARY = 1
-BINARY = 2
+
+__all__ = [
+    'PRODUCT',
+    'DIFFERENCE',
+    'UNION',
+    'INTERSECTION',
+    'DIVISION',
+    'JOIN',
+    'JOIN_LEFT',
+    'JOIN_RIGHT',
+    'JOIN_FULL',
+    'PROJECTION',
+    'SELECTION',
+    'RENAME',
+    'ARROW',
+    'TokenizerException',
+    'ParserException',
+    'CallableString',
+    'Node',
+    'Unary',
+    'Binary',
+    'Variable',
+    'tree',
+    'parse',
+]
 
 PRODUCT = '*'
 DIFFERENCE = '-'
@@ -44,6 +69,7 @@ SELECTION = 'σ'
 RENAME = 'ρ'
 ARROW = '➡'
 
+
 b_operators = (PRODUCT, DIFFERENCE, UNION, INTERSECTION, DIVISION,
                JOIN, JOIN_LEFT, JOIN_RIGHT, JOIN_FULL)  # List of binary operators
 u_operators = (PROJECTION, SELECTION, RENAME)  # List of unary operators
@@ -82,9 +108,8 @@ class CallableString(str):
         '''
         return eval(self, context)
 
-
-class Node (object):
-
+@dataclass
+class Node:
     '''This class is a node of a relational expression. Leaves are relations
     and internal nodes are operations.
 
@@ -100,77 +125,17 @@ class Node (object):
     operation.
 
     This class is used to convert an expression into python code.'''
-    kind = None
-    __hash__ = None
+    name: str
 
-    def __init__(self, expression=None):
-        '''Generates the tree from the tokenized expression
-        If no expression is specified then it will create an empty node'''
-        if expression == None or len(expression) == 0:
-            return
+    def __init__(self, name: str) -> None:
+        raise NotImplementedError('This is supposed to be an abstract class')
 
-        # If the list contains only a list, it will consider the lower level list.
-        # This will allow things like ((((((a))))) to work
-        while len(expression) == 1 and isinstance(expression[0], list):
-            expression = expression[0]
-
-        # The list contains only 1 string. Means it is the name of a relation
-        if len(expression) == 1:
-            self.kind = RELATION
-            self.name = expression[0]
-            if not rtypes.is_valid_relation_name(self.name):
-                raise ParserException(
-                    u"'%s' is not a valid relation name" % self.name)
-            return
-
-        # Expression from right to left, searching for binary operators
-        # this means that binary operators have lesser priority than
-        # unary operators.
-        # It finds the operator with lesser priority, uses it as root of this
-        # (sub)tree using everything on its left as left parameter (so building
-        # a left subtree with the part of the list located on left) and doing
-        # the same on right.
-        # Since it searches for strings, and expressions into parenthesis are
-        # within sub-lists, they won't be found here, ensuring that they will
-        # have highest priority.
-        for i in range(len(expression) - 1, -1, -1):
-            if expression[i] in b_operators:  # Binary operator
-                self.kind = BINARY
-                self.name = expression[i]
-
-                if len(expression[:i]) == 0:
-                    raise ParserException(
-                        u"Expected left operand for '%s'" % self.name)
-
-                if len(expression[i + 1:]) == 0:
-                    raise ParserException(
-                        u"Expected right operand for '%s'" % self.name)
-
-                self.left = node(expression[:i])
-                self.right = node(expression[i + 1:])
-                return
-        '''Searches for unary operators, parsing from right to left'''
-        for i in range(len(expression) - 1, -1, -1):
-            if expression[i] in u_operators:  # Unary operator
-                self.kind = UNARY
-                self.name = expression[i]
-
-                if len(expression) <= i + 2:
-                    raise ParserException(
-                        u"Expected more tokens in '%s'" % self.name)
-
-                self.prop = expression[1 + i].strip()
-                self.child = node(expression[2 + i])
-
-                return
-        raise ParserException("Expected operator in '%s'" % expression)
-
-    def toCode(self):
+    def toCode(self): #FIXME return type
         '''This method converts the AST into a python code object'''
         code = self._toPython()
         return compile(code, '<relational_expression>', 'eval')
 
-    def toPython(self):
+    def toPython(self) -> CallableString:
         '''This method converts the AST into a python code string, which
         will require the relation module to be executed.
 
@@ -178,88 +143,57 @@ class Node (object):
         directly called.'''
         return CallableString(self._toPython())
 
-    def _toPython(self):
-        '''
-        Same as toPython but returns a regular string
-        '''
-        if self.name in b_operators:
-            return '%s.%s(%s)' % (self.left.toPython(), op_functions[self.name], self.right.toPython())
-        elif self.name in u_operators:
-            prop = self.prop
+    def _toPython(self) -> str:
+        raise NotImplementedError()
 
-            # Converting parameters
-            if self.name == PROJECTION:
-                prop = '\"%s\"' % prop.replace(' ', '').replace(',', '\",\"')
-            elif self.name == RENAME:
-                prop = '{\"%s\"}' % prop.replace(
-                    ',', '\",\"').replace(ARROW, '\":\"').replace(' ', '')
-            else:  # Selection
-                prop = repr(prop)
-
-            return '%s.%s(%s)' % (self.child.toPython(), op_functions[self.name], prop)
-        return self.name
-
-    def printtree(self, level=0):
+    def printtree(self, level: int = 0) -> str:
         '''returns a representation of the tree using indentation'''
-        r = ''
-        for i in range(level):
-            r += '  '
-        r += self.name
-        if self.name in b_operators:
+        r = '  ' * level + self.name
+        if self.name in b_operators and isinstance(self, Binary):
             r += self.left.printtree(level + 1)
             r += self.right.printtree(level + 1)
-        elif self.name in u_operators:
+        elif self.name in u_operators and isinstance(self, Unary):
             r += '\t%s\n' % self.prop
             r += self.child.printtree(level + 1)
-
         return '\n' + r
 
-    def get_left_leaf(self):
-        '''This function returns the leftmost leaf in the tree.'''
-        if self.kind == RELATION:
-            return self
-        elif self.kind == UNARY:
-            return self.child.get_left_leaf()
-        elif self.kind == BINARY:
-            return self.left.get_left_leaf()
+    def get_left_leaf(self) -> 'Node':
+        raise NotImplementedError()
 
-    def result_format(self, rels):
+    def result_format(self, rels: dict) -> list: #FIXME types
         '''This function returns a list containing the fields that the resulting relation will have.
         It requires a dictionary where keys are the names of the relations and the values are
         the relation objects.'''
-        if rels == None:
-            return
+        if not isinstance(rels, dict):
+            raise TypeError('Can\'t be of None type')
 
-        if self.kind == RELATION:
+        if isinstance(self, Variable):  #FIXME this is ugly
             return list(rels[self.name].header)
-        elif self.kind == BINARY and self.name in (DIFFERENCE, UNION, INTERSECTION):
-            return self.left.result_format(rels)
-        elif self.kind == BINARY and self.name == DIVISION:
-            return list(set(self.left.result_format(rels)) - set(self.right.result_format(rels)))
-        elif self.name == PROJECTION:
-            l = []
-            for i in self.prop.split(','):
-                l.append(i.strip())
-            return l
-        elif self.name == PRODUCT:
-            return self.left.result_format(rels) + self.right.result_format(rels)
-        elif self.name == SELECTION:
-            return self.child.result_format(rels)
-        elif self.name == RENAME:
-            _vars = {}
-            for i in self.prop.split(','):
-                q = i.split(ARROW)
-                _vars[q[0].strip()] = q[1].strip()
+        elif isinstance(self, Binary):
+            if self.name in (DIFFERENCE, UNION, INTERSECTION):
+                return self.left.result_format(rels)
+            elif self.name == DIVISION:
+                return list(set(self.left.result_format(rels)) - set(self.right.result_format(rels)))
+            elif self.name == PRODUCT:
+                return self.left.result_format(rels) + self.right.result_format(rels)
+            elif self.name in (JOIN, JOIN_LEFT, JOIN_RIGHT, JOIN_FULL):
+                return list(set(self.left.result_format(rels)).union(set(self.right.result_format(rels))))
+        elif isinstance(self, Unary):
+            if self.name == PROJECTION:
+                return self.get_projection_prop()
+            elif self.name == SELECTION:
+                return self.child.result_format(rels)
+            elif self.name == RENAME:
+                _vars = self.get_rename_prop()
+                _fields = self.child.result_format(rels)
+                for i in range(len(_fields)):
+                    if _fields[i] in _vars:
+                        _fields[i] = _vars[_fields[i]]
+                return _fields
 
-            _fields = self.child.result_format(rels)
-            for i in range(len(_fields)):
-                if _fields[i] in _vars:
-                    _fields[i] = _vars[_fields[i]]
-            return _fields
-        elif self.name in (JOIN, JOIN_LEFT, JOIN_RIGHT, JOIN_FULL):
-            return list(set(self.left.result_format(rels)).union(set(self.right.result_format(rels))))
+        raise ValueError('What kind of alien object is this?')
 
-    def __eq__(self, other):
+    def __eq__(self, other): #FIXME
         if not (isinstance(other, node) and self.name == other.name and self.kind == other.kind):
             return False
 
@@ -271,22 +205,156 @@ class Node (object):
             return self.left == other.left and self.right == other.right
         return True
 
+
+@dataclass
+class Variable(Node):
+    def _toPython(self) -> str:
+        return self.name
+
     def __str__(self):
-        if (self.kind == RELATION):
-            return self.name
-        elif (self.kind == UNARY):
-            return self.name + " " + self.prop + " (" + self.child.__str__() + ")"
-        elif (self.kind == BINARY):
-            le = self.left.__str__()
-            if self.right.kind != BINARY:
-                re = self.right.__str__()
-            else:
-                re = "(" + self.right.__str__() + ")"
+        return self.name
 
-            return (le + self.name + re)
+    def get_left_leaf(self) -> Node:
+        return self
 
 
-def _find_matching_parenthesis(expression, start=0, openpar=u'(', closepar=u')'):
+@dataclass
+class Binary(Node):
+    name: str
+    left: Node
+    right: Node
+
+    def get_left_leaf(self) -> Node:
+        return self.left.get_left_leaf()
+
+    def _toPython(self) -> str:
+        return '%s.%s(%s)' % (self.left._toPython(), op_functions[self.name], self.right._toPython())
+
+    def __str__(self):
+        le = self.left.__str__()
+        if isinstance(self.right, Binary):
+            re = "(" + self.right.__str__() + ")"
+        else:
+            re = self.right.__str__()
+        return (le + self.name + re) #TODO use fstrings
+
+
+@dataclass
+class Unary(Node):
+    name: str
+    prop: str
+    child: Node
+
+    def get_left_leaf(self) -> Node:
+        return self.child.get_left_leaf()
+
+    def __str__(self):
+        return self.name + " " + self.prop + " (" + self.child.__str__() + ")" #TODO use fstrings
+
+    def _toPython(self) -> str:
+        prop = self.prop
+
+        # Converting parameters
+        if self.name == PROJECTION:
+            prop = repr(self.get_projection_prop())
+        elif self.name == RENAME:
+            prop = repr(self.get_rename_prop())
+        else:  # Selection
+            prop = repr(prop)
+
+        return '%s.%s(%s)' % (self.child._toPython(), op_functions[self.name], prop)
+
+    def get_projection_prop(self) -> List[str]:
+        if self.name != PROJECTION:
+            raise ValueError('This is only supported on projection nodes')
+        return [i.strip() for i in self.prop.split(',')]
+
+    def set_projection_prop(self, p: List[str]) -> None:
+        if self.name != PROJECTION:
+            raise ValueError('This is only supported on projection nodes')
+        self.prop = ','.join(p)
+
+    def get_rename_prop(self) -> Dict[str, str]:
+        '''
+        Returns the dictionary that the rename operation wants
+        '''
+        if self.name != RENAME:
+            raise ValueError('This is only supported on rename nodes')
+        r = {}
+        for i in self.prop.split(','):
+            q = i.split(ARROW)
+            r[q[0].strip()] = q[1].strip()
+        return r
+
+    def set_rename_prop(self, renames: Dict[str, str]) -> None:
+        '''
+        Sets the prop field based on the dictionary for renames
+        '''
+        if self.name != RENAME:
+            raise ValueError('This is only supported on rename nodes')
+        self.prop = ','.join(f'{k}{ARROW}{v}' for k, v in renames.items())
+
+
+def parse_tokens(expression: List[Union[list, str]]) -> Node:
+    '''Generates the tree from the tokenized expression
+    If no expression is specified then it will create an empty node'''
+
+    # If the list contains only a list, it will consider the lower level list.
+    # This will allow things like ((((((a))))) to work
+    while len(expression) == 1 and isinstance(expression[0], list):
+        expression = expression[0]
+
+    if len(expression) == 0:
+        raise ParserException('Failed to parse empty expression')
+
+    # The list contains only 1 string. Means it is the name of a relation
+    if len(expression) == 1:
+        assert isinstance(expression[0], str)
+        if not rtypes.is_valid_relation_name(expression[0]):
+            raise ParserException(
+                f'{expression[0]!r} is not a valid relation name')
+        return Variable(expression[0]) #FIXME Move validation in the object
+
+    # Expression from right to left, searching for binary operators
+    # this means that binary operators have lesser priority than
+    # unary operators.
+    # It finds the operator with lesser priority, uses it as root of this
+    # (sub)tree using everything on its left as left parameter (so building
+    # a left subtree with the part of the list located on left) and doing
+    # the same on right.
+    # Since it searches for strings, and expressions into parenthesis are
+    # within sub-lists, they won't be found here, ensuring that they will
+    # have highest priority.
+    for i in range(len(expression) - 1, -1, -1):
+        if expression[i] in b_operators:  # Binary operator
+
+            if len(expression[:i]) == 0:
+                raise ParserException(
+                    f'Expected left operand for {expression[i]!r}')
+
+            if len(expression[i + 1:]) == 0:
+                raise ParserException(
+                    f'Expected right operand for {expression[i]!r}')
+            return Binary(expression[i], parse_tokens(expression[:i]), parse_tokens(expression[i + 1:]))  # type: ignore
+    '''Searches for unary operators, parsing from right to left'''
+    for i in range(len(expression)):
+        if expression[i] in u_operators:  # Unary operator
+            if len(expression) <= i + 2:
+                raise ParserException(
+                    f'Expected more tokens in {expression[i]!r}')
+            elif len(expression) > i + 3:
+                raise ParserException(
+                    f'Too many tokens in {expression[i]!r}')
+
+            return Unary(
+                expression[i],  # type: ignore
+                prop=expression[1 + i].strip(),  # type: ignore
+                child=parse_tokens(expression[2 + i])  # type: ignore
+            )
+    raise ParserException(f'Parse error on {expression!r}')
+
+
+def _find_matching_parenthesis(expression: str, start=0, openpar='(', closepar=')') -> Optional[int]:
     '''This function returns the position of the matching
     close parenthesis to the 1st open parenthesis found
     starting from start (0 by default)'''
@@ -305,15 +373,15 @@ def _find_matching_parenthesis(expression, start=0, openpar=u'(', closepar=u')')
         if string:
             continue
 
-
         if expression[i] == openpar:
             par_count += 1
         elif expression[i] == closepar:
             par_count -= 1
             if par_count == 0:
                 return i  # Closing parenthesis of the parameter
+    return None
 
-def _find_token(haystack, needle):
+def _find_token(haystack: str, needle: str) -> int:
     '''
     Like the string function find, but
     ignores tokens that are within a string
@@ -338,20 +406,20 @@ def _find_token(haystack, needle):
     return r
 
 
-def tokenize(expression):
+def tokenize(expression: str) -> list:
     '''This function converts a relational expression into a list where
     every token of the expression is an item of a list. Expressions into
     parenthesis will be converted into sublists.'''
 
-    items = []  # List for the tokens
+    # List for the tokens
+    items = [] #  type: List[Union[str,list]]
 
     expression = expression.strip()  # Removes initial and ending spaces
 
     while len(expression) > 0:
-
         if expression.startswith('('):  # Parenthesis state
             end = _find_matching_parenthesis(expression)
-            if end == None:
+            if end is None:
                 raise TokenizerException(
                     "Missing matching ')' in '%s'" % expression)
             # Appends the tokenization of the content of the parenthesis
@@ -385,27 +453,18 @@ def tokenize(expression):
                     break
             items.append(expression[:r])
             expression = expression[r:].strip()
-
     return items
 
 
-def tree(expression):
+def tree(expression: str) -> Node:
     '''This function parses a relational algebra expression into a AST and returns
     the root node using the Node class.'''
-    return node(tokenize(expression))
+    return parse_tokens(tokenize(expression))
 
 
-def parse(expr):
+def parse(expr: str) -> CallableString:
     '''This function parses a relational algebra expression, and returns a
     CallableString (a string that can be called) whith the corresponding
     Python expression.
     '''
     return tree(expr).toPython()
-
-if __name__ == "__main__":
-    while True:
-        e = input("Expression: ")
-        print (parse(e))
-
-# Backwards compatibility
-node = Node
