@@ -1,5 +1,5 @@
 # Relational
-# Copyright (C) 2008-2016  Salvo "LtWorf" Tomaselli
+# Copyright (C) 2008-2020  Salvo "LtWorf" Tomaselli
 #
 # Relational is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -19,15 +19,22 @@
 # This module provides a classes to represent relations and to perform
 # relational operations on them.
 
-import csv
-from itertools import chain, repeat
+from itertools import chain, repeat, product as iproduct
 from collections import deque
+from typing import FrozenSet, Iterable, List, Dict, Tuple, Optional
+from dataclasses import dataclass
+from pathlib import Path
 
 from relational.rtypes import *
 
 
-class Relation (object):
+__all__ = [
+    'Relation',
+    'Header',
+]
 
+@dataclass(repr=True, unsafe_hash=False, frozen=True)
+class Relation:
     '''
     This object defines a relation (as a group of consistent tuples) and operations.
 
@@ -50,54 +57,63 @@ class Relation (object):
     An empty relation needs a header, and can be filled using the insert()
     method.
     '''
-    __hash__ = None
+    header: 'Header'
+    content: FrozenSet[Tuple[CastValue, ...]]
 
-    def __init__(self, filename=""):
-        self._readonly = False
-        self.content = set()
+    @staticmethod
+    def load_csv(filename: Union[str, Path]) -> 'Relation':
+        '''
+        Load a relation object from a csv file.
 
-        if len(filename) == 0:  # Empty relation
-            self.header = Header([])
-            return
+        The 1st row is the header and the other rows are the content.
+
+        Types will be inferred automatically
+        '''
+        import csv
         with open(filename) as fp:
             reader = csv.reader(fp)  # Creating a csv reader
-            self.header = Header(next(reader))  # read 1st line
-            iterator = ((self.insert(i) for i in reader))
-            deque(iterator, maxlen=0)
+            header = Header(next(reader))  # read 1st line
+            return Relation.create_from(header, reader)
 
-    def _make_duplicate(self, copy):
-        '''Flag that the relation "copy" is pointing
-        to the same set as this relation.'''
+    @staticmethod
+    def load(filename: Union[str, Path]) -> 'Relation':
+        '''
+        Load a relation object from a json file.
+        '''
+        with open(filename) as fp:
+            from json import load as jload
+            from typedload import load
+            loaded = jload(fp)
+            header = Header(loaded['header'])
+            content = []
+            for row in loaded['content']:
+                if len(row) != len(header):
+                    raise ValueError(f'Line {row} contains an incorrect amount of values')
+                t_row: Tuple[Optional[Union[int, float, str, Rdate]], ...] = load(
+                    row,
+                    Tuple[Optional[Union[int, float, str, Rdate]], ...],  # type: ignore
+                    basiccast=False
+                )
+                content.append(t_row)
+            return Relation(header, frozenset(content))
 
-        self._readonly = True
-        copy._readonly = True
+    def save(self, filename: Union[Path, str]) -> None:
+        '''
+        Saves the relation in a file.
+        Will save using the json format
+        '''
+        with open(filename, 'w') as fp:
+            from json import dump as jdump
+            from typedload import dump
+            jdump(dump(self), fp)
 
-    def _make_writable(self, copy_content=True):
-        '''If this relation is marked as readonly, this
-        method will copy the content to make it writable too
-
-        if copy_content is set to false, the caller must
-        separately copy the content.'''
-
-        if self._readonly:
-            self._readonly = False
-
-            if copy_content:
-                self.content = set(self.content)
-
-    def __iter__(self):
-        return iter(self.content)
-
-    def __contains__(self, key):
-        return key in self.content
-
-    def save(self, filename):
+    def save_csv(self, filename: Union[Path, str]) -> None:
         '''
         Saves the relation in a file. Will save using the csv
         format as defined in RFC4180.
         '''
-
-        with open(filename, 'w') as fp:
+        import csv
+        with open(filename, 'w', newline='\n') as fp:
             writer = csv.writer(fp)  # Creating csv writer
 
             # It wants an iterable containing iterables
@@ -107,13 +123,47 @@ class Relation (object):
             # Writing content, already in the correct format
             writer.writerows(self.content)
 
-    def _rearrange(self, other):
+    @staticmethod
+    def create_from(header: Iterable[str], content: Iterable[List[str]]) -> 'Relation':
+        '''
+        Iterator for the header, and iterator for the content.
+
+        This will infer types.
+        '''
+        header = Header(header)
+        r_content = []
+        guessed_types = list(repeat({Rdate, float, int, str}, len(header)))
+
+        for row in content:
+            if len(row) != len(header):
+                raise ValueError(f'Line {row} contains an incorrect amount of values')
+            r_content.append(row)
+
+            # Guess types
+            for i, value in enumerate(row):
+                guessed_types[i] = guessed_types[i].intersection(guess_type(value))
+
+        typed_content = []
+        for r in r_content:
+            t = tuple(cast(v, guessed_types[i]) for i, v in enumerate(r))
+            typed_content.append(t)
+
+        return Relation(header, frozenset(typed_content))
+
+
+    def __iter__(self):
+        return iter(self.content)
+
+    def __contains__(self, key):
+        return key in self.content
+
+    def _rearrange(self, other: 'Relation') -> 'Relation':
         '''If two relations share the same attributes in a different order, this method
         will use projection to make them have the same attributes' order.
         It is not exactely related to relational algebra. Just a method used
         internally.
         Will raise an exception if they don't share the same attributes'''
-        if not isinstance(other, relation):
+        if not isinstance(other, Relation):
             raise TypeError('Expected an instance of the same class')
         elif self.header == other.header:
             return other
@@ -123,52 +173,46 @@ class Relation (object):
             ','.join(self.header), ','.join(other.header)
         ))
 
-    def selection(self, expr):
+    def selection(self, expr: str) -> 'Relation':
         '''
         Selection, expr must be a valid Python expression; can contain field names.
         '''
-        newt = relation()
-        newt.header = Header(self.header)
-
         try:
             c_expr = compile(expr, 'selection', 'eval')
         except:
-            raise Exception('Failed to compile expression: %s' % expr)
+            raise Exception(f'Failed to compile expression: {expr}')
 
+        content = []
         for i in self.content:
             # Fills the attributes dictionary with the values of the tuple
-            attributes = {attr: i[j].autocast()
+            attributes = {attr: i[j]
                           for j, attr in enumerate(self.header)
                           }
 
             try:
                 if eval(c_expr, attributes):
-                    newt.content.add(i)
+                    content.append(i)
             except Exception as e:
-                raise Exception(
-                    "Failed to evaluate %s\n%s" % (expr, e.__str__()))
-        return newt
+                raise Exception(f'Failed to evaluate {expr} with {i}\n{e}')
+        return Relation(self.header, frozenset(content))
 
-    def product(self, other):
+    def product(self, other: 'Relation') -> 'Relation':
         '''
         Cartesian product. Attributes of the relations must differ.
         '''
 
-        if (not isinstance(other, relation)):
+        if (not isinstance(other, Relation)):
             raise Exception('Operand must be a relation')
         if self.header.sharedAttributes(other.header) != 0:
             raise Exception(
                 'Unable to perform product on relations with colliding attributes'
             )
-        newt = relation()
-        newt.header = Header(self.header + other.header)
+        header = Header(self.header + other.header)
 
-        for i in self.content:
-            for j in other.content:
-                newt.content.add(i + j)
-        return newt
+        content = frozenset(i+j for i, j in iproduct(self.content, other.content))
+        return Relation(header, content)
 
-    def projection(self, * attributes):
+    def projection(self, *attributes) -> 'Relation':
         '''
         Can be called in two different ways:
         a.projection('field1','field2')
@@ -188,18 +232,13 @@ class Relation (object):
 
         if len(ids) == 0:
             raise Exception('Invalid attributes for projection')
-        newt = relation()
-        # Create the header
-        h = (self.header[i] for i in ids)
-        newt.header = Header(h)
+        header = Header((self.header[i] for i in ids))
 
-        # Create the body
-        for i in self.content:
-            row = (i[j] for j in ids)
-            newt.content.add(tuple(row))
-        return newt
+        content = frozenset(tuple((i[j] for j in ids)) for i in self.content)
 
-    def rename(self, params):
+        return Relation(header, content)
+
+    def rename(self, params: Dict[str, str]) -> 'Relation':
         '''
         Takes a dictionary.
 
@@ -208,38 +247,26 @@ class Relation (object):
         For example if you want to rename a to b, call
         rel.rename({'a':'b'})
         '''
-        newt = relation()
-        newt.header = self.header.rename(params)
+        header = self.header.rename(params)
+        return Relation(header, self.content)
 
-        newt.content = self.content
-        self._make_duplicate(newt)
-        return newt
-
-    def intersection(self, other):
+    def intersection(self, other: 'Relation') -> 'Relation':
         '''
         Intersection operation. The result will contain items present in both
         operands.
         Will return an empty one if there are no common items.
         '''
         other = self._rearrange(other)  # Rearranges attributes' order
-        newt = relation()
-        newt.header = Header(self.header)
+        return Relation(self.header, self.content.intersection(other.content))
 
-        newt.content = self.content.intersection(other.content)
-        return newt
-
-    def difference(self, other):
+    def difference(self, other: 'Relation') -> 'Relation':
         '''Difference operation. The result will contain items present in first
         operand but not in second one.
         '''
         other = self._rearrange(other)  # Rearranges attributes' order
-        newt = relation()
-        newt.header = Header(self.header)
+        return Relation(self.header, self.content.difference(other.content))
 
-        newt.content = self.content.difference(other.content)
-        return newt
-
-    def division(self, other):
+    def division(self, other: 'Relation') -> 'Relation':
         '''Division operator
         The division is a binary operation that is written as R ÷ S. The
         result consists of the restrictions of tuples in R to the
@@ -265,50 +292,42 @@ class Relation (object):
         t = self.projection(d_headers).product(other)
         return self.projection(d_headers).difference(t.difference(self).projection(d_headers))
 
-    def union(self, other):
+    def union(self, other: 'Relation') -> 'Relation':
         '''Union operation. The result will contain items present in first
         and second operands.
         '''
         other = self._rearrange(other)  # Rearranges attributes' order
-        newt = relation()
-        newt.header = Header(self.header)
+        return Relation(self.header, self.content.union(other.content))
 
-        newt.content = self.content.union(other.content)
-        return newt
-
-    def thetajoin(self, other, expr):
+    def thetajoin(self, other: 'Relation', expr: str) -> 'Relation':
         '''Defined as product and then selection with the given expression.'''
         return self.product(other).selection(expr)
 
-    def outer(self, other):
+    def outer(self, other: 'Relation') -> 'Relation':
         '''Does a left and a right outer join and returns their union.'''
         a = self.outer_right(other)
         b = self.outer_left(other)
 
         return a.union(b)
 
-    def outer_right(self, other):
+    def outer_right(self, other: 'Relation') -> 'Relation':
         '''
         Outer right join. Considers self as left and param as right. If the
-        tuple has no corrispondence, empy attributes are filled with a "---"
-        string. This is due to the fact that the None token would cause
-        problems when saving and reloading the relation.
+        tuple has no corrispondence, empy attributes are filled with a None.
         Just like natural join, it works considering shared attributes.
         '''
         return other.outer_left(self)
 
-    def outer_left(self, other, swap=False):
+    def outer_left(self, other: 'Relation', swap=False) -> 'Relation':
         '''
         See documentation for outer_right
         '''
-
         shared = self.header.intersection(other.header)
 
-        newt = relation()  # Creates the new relation
         # Creating the header with all the fields, done like that because order is
         # needed
         h = (i for i in other.header if i not in shared)
-        newt.header = Header(chain(self.header, h))
+        header = Header(chain(self.header, h))
 
         # Shared ids of self
         sid = self.header.getAttributesId(shared)
@@ -318,6 +337,7 @@ class Relation (object):
         # Non shared ids of the other relation
         noid = [i for i in range(len(other.header)) if i not in oid]
 
+        content = []
         for i in self.content:
             # Tuple partecipated to the join?
             added = False
@@ -329,16 +349,16 @@ class Relation (object):
                 if match:
                     item = chain(i, (j[l] for l in noid))
 
-                    newt.content.add(tuple(item))
+                    content.append(tuple(item))
                     added = True
             # If it didn't partecipate, adds it
             if not added:
-                item = chain(i, repeat(rstring('---'), len(noid)))
-                newt.content.add(tuple(item))
+                item = chain(i, repeat(None, len(noid)))
+                content.append(tuple(item))
 
-        return newt
+        return Relation(header, frozenset(content))
 
-    def join(self, other):
+    def join(self, other: 'Relation') -> 'Relation':
         '''
         Natural join, joins on shared attributes (one or more). If there are no
         shared attributes, it will behave as the cartesian product.
@@ -347,12 +367,10 @@ class Relation (object):
         # List of attributes in common between the relations
         shared = self.header.intersection(other.header)
 
-        newt = relation()  # Creates the new relation
-
         # Creating the header with all the fields, done like that because order is
         # needed
         h = (i for i in other.header if i not in shared)
-        newt.header = Header(chain(self.header, h))
+        header = Header(chain(self.header, h))
 
         # Shared ids of self
         sid = self.header.getAttributesId(shared)
@@ -362,6 +380,7 @@ class Relation (object):
         # Non shared ids of the other relation
         noid = [i for i in range(len(other.header)) if i not in oid]
 
+        content = []
         for i in self.content:
             for j in other.content:
                 match = True
@@ -370,12 +389,12 @@ class Relation (object):
 
                 if match:
                     item = chain(i, (j[l] for l in noid))
-                    newt.content.add(tuple(item))
+                    content.append(tuple(item))
 
-        return newt
+        return Relation(header, frozenset(content))
 
     def __eq__(self, other):
-        if not isinstance(other, relation):
+        if not isinstance(other, Relation):
             return False
 
         if len(self.content) != len(other.content):
@@ -393,95 +412,52 @@ class Relation (object):
     def __len__(self):
         return len(self.content)
 
-    def __str__(self):
-        m_len = [len(i) for i in self.header]  # Maximum lenght string
+    def __str__(self) -> str:
+        return self.pretty_string(tty=False)
+
+    def pretty_string(self, tty: bool) -> str:
+        '''
+        Returns a printable string.
+
+        If tty is enabled, it will attempt to add ANSI color codes
+        '''
+        c = lambda i, ansi: i
+        if not tty:
+            colorize = c
+        else:
+            from os import isatty
+            if isatty(1) and isatty(2):
+                try:
+                    from xtermcolor import colorize  # type: ignore
+                except ModuleNotFoundError:
+                    colorize = c
+            else:
+                colorize = c
+        m_len = [len(i) + 2 for i in self.header]  # Maximum lenght string
 
         for f in self.content:
-            for col, i in enumerate(f):
-                if len(i) > m_len[col]:
-                    m_len[col] = len(i)
+            for col, k in enumerate(str(val) for val in f):
+                if len(k) + 2 > m_len[col]:
+                    m_len[col] = len(k) + 2
 
         res = ""
-        for f, attr in enumerate(self.header):
-            res += "%s" % (attr.ljust(2 + m_len[f]))
+        for j, attr in enumerate(self.header):
+            res += colorize(attr.ljust(m_len[j]), ansi=3)
 
         for r in self.content:
             res += "\n"
             for col, i in enumerate(r):
-                res += "%s" % (i.ljust(2 + m_len[col]))
+                cell = str(i).ljust(m_len[col])
+                if isinstance(i, (int, float)):
+                    cell = colorize(cell, ansi=4)
+                elif i is None:
+                    cell = colorize(cell, ansi=1)
+                elif not isinstance(i, str):
+                    cell = colorize(cell, ansi=15)
+                res += cell
 
         return res
 
-    def update(self, expr, dic):
-        '''
-        Updates certain values of a relation.
-
-        expr must be a valid Python expression that can contain field names.
-
-        This operation will change the relation itself instead of generating a new one,
-        updating all the tuples where expr evaluates as True.
-
-        Dic must be a dictionary that has the form "field name":"new value". Every kind of value
-        will be converted into a string.
-
-        Returns the number of affected rows.
-        '''
-        self._make_writable(copy_content=False)
-        affected = self.selection(expr)
-        not_affected = self.difference(affected)
-
-        new_values = tuple(
-            zip(self.header.getAttributesId(dic.keys()), dic.values())
-        )
-
-        for i in set(affected.content):
-            i = list(i)
-
-            for column, value in new_values:
-                i[column] = value
-            not_affected.insert(i)
-
-        self.content = not_affected.content
-        return len(affected)
-
-    def insert(self, values):
-        '''
-        Inserts a tuple in the relation.
-        This function will not insert duplicate tuples.
-        All the values will be converted in string.
-        Will return the number of inserted rows.
-
-        Will fail if the tuple has the wrong amount of items.
-        '''
-
-        if len(self.header) != len(values):
-            raise Exception(
-                'Tuple has the wrong size. Expected %d, got %d' % (
-                    len(self.header),
-                    len(values)
-                )
-            )
-
-        self._make_writable()
-
-        prevlen = len(self.content)
-        self.content.add(tuple(map(rstring, values)))
-        return len(self.content) - prevlen
-
-    def delete(self, expr):
-        '''
-        Delete, expr must be a valid Python expression; can contain field names.
-
-        This operation will change the relation itself instead of generating a new one,
-        deleting all the tuples where expr evaluates as True.
-
-        Returns the number of affected rows.'''
-
-        l = len(self.content)
-        self._make_writable(copy_content=False)
-        self.content = self.difference(self.selection(expr)).content
-        return len(self.content) - l
-
 
 class Header(tuple):
 
@@ -496,7 +472,7 @@ class Header(tuple):
 
         for i in self:
             if not is_valid_relation_name(i):
-                raise Exception('"%s" is not a valid attribute name' % i)
+                raise Exception(f'"{i}" is not a valid attribute name')
 
         if len(self) != len(set(self)):
             raise Exception('Attribute names must be unique')
@@ -504,7 +480,7 @@ class Header(tuple):
     def __repr__(self):
         return "Header(%s)" % super(Header, self).__repr__()
 
-    def rename(self, params):
+    def rename(self, params: Dict[str, str]) -> 'Header':
         '''Returns a new header, with renamed fields.
 
         params is a dictionary of {old:new} names
@@ -512,33 +488,29 @@ class Header(tuple):
         attrs = list(self)
         for old, new in params.items():
             if not is_valid_relation_name(new):
-                raise Exception('%s is not a valid attribute name' % new)
+                raise Exception(f'{new} is not a valid attribute name')
             try:
                 id_ = attrs.index(old)
                 attrs[id_] = new
             except:
-                raise Exception('Field not found: %s' % old)
+                raise Exception(f'Field not found: {old}')
         return Header(attrs)
 
-    def sharedAttributes(self, other):
+    def sharedAttributes(self, other: 'Header') -> int:
         '''Returns how many attributes this header has in common with a given one'''
         return len(set(self).intersection(set(other)))
 
-    def union(self, other):
+    def union(self, other: 'Header') -> Set[str]:
         '''Returns the union of the sets of attributes with another header.'''
         return set(self).union(set(other))
 
-    def intersection(self, other):
+    def intersection(self, other: 'Header') -> Set[str]:
         '''Returns the set of common attributes with another header.'''
         return set(self).intersection(set(other))
 
-    def getAttributesId(self, param):
+    def getAttributesId(self, param: Iterable[str]) -> List[int]:
         '''Returns a list with numeric index corresponding to field's name'''
         try:
             return [self.index(i) for i in param]
         except ValueError as e:
             raise Exception('One of the fields is not in the relation: %s' % ','.join(param))
-
-# Backwards compatibility
-relation = Relation
-header = Header
