@@ -1,5 +1,5 @@
 # Relational
-# Copyright (C) 2008-2015  Salvo "LtWorf" Tomaselli
+# Copyright (C) 2008-2020  Salvo "LtWorf" Tomaselli
 #
 # Relational is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -35,7 +35,7 @@ class creatorForm(QtWidgets.QDialog):
         self.ui = ui
         self.table = self.ui.table
 
-        if self.rel == None:
+        if self.rel is None:
             self.setup_empty()
         else:
             self.setup_relation(self.rel)
@@ -52,13 +52,13 @@ class creatorForm(QtWidgets.QDialog):
 
         for i in rel.content:
             self.table.insertRow(self.table.rowCount())
-            for j in range(len(i)):
+            for j, value in enumerate(i):
+                if value is None:
+                    raise Exception('Relation contains a None value and cannot be edited from the GUI')
                 item = QtWidgets.QTableWidgetItem()
-                item.setText(i[j])
+                item.setText(str(value))
                 self.table.setItem(self.table.rowCount() - 1, j, item)
 
-        pass
-
     def setup_empty(self):
         self.table.insertColumn(0)
         self.table.insertColumn(0)
@@ -84,14 +84,13 @@ class creatorForm(QtWidgets.QDialog):
              for i in range(self.table.columnCount()))
 
         try:
-            header = relation.header(h)
+            header = relation.Header(h)
         except Exception as e:
             QtWidgets.QMessageBox.information(None, QtWidgets.QApplication.translate("Form", "Error"), "%s\n%s" % (
                 QtWidgets.QApplication.translate("Form", "Header error!"), e.__str__()))
             return None
-        r = relation.relation()
-        r.header = header
 
+        content = []
         for i in range(1, self.table.rowCount()):
             hlist = []
             for j in range(self.table.columnCount()):
@@ -101,11 +100,10 @@ class creatorForm(QtWidgets.QDialog):
                     QtWidgets.QMessageBox.information(None, QtWidgets.QApplication.translate(
                         "Form", "Error"), QtWidgets.QApplication.translate("Form", "Unset value in %d,%d!" % (i + 1, j + 1)))
                     return None
-            r.insert(hlist)
-        return r
+            content.append(hlist)
+        return relation.Relation.create_from(header, content)
 
     def accept(self):
-
         self.result_relation = self.create_relation()
 
         # Doesn't close the window in case of errors
@@ -144,11 +142,3 @@ def edit_relation(rel=None):
 
     Form.exec_()
     return Form.result_relation
-
-
-if __name__ == '__main__':
-    import sys
-    app = QtGui.QApplication(sys.argv)
-    r = relation.relation(
-        "/home/salvo/dev/relational/trunk/samples/people.csv")
-    print (edit_relation(r))
