@@ -1,5 +1,5 @@
 # Relational
-# Copyright (C) 2008-2016  Salvo "LtWorf" Tomaselli
+# Copyright (C) 2008-2020  Salvo "LtWorf" Tomaselli
 #
 # Relational is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -28,6 +28,9 @@ from relational_gui import surveyForm
 from relational_gui import maingui
 
 
+version = ''
+
+
 class relForm(QtWidgets.QMainWindow):
 
     def __init__(self):
@@ -236,7 +239,7 @@ class relForm(QtWidgets.QMainWindow):
         '''Shows the selected relation into the table'''
         self.ui.table.clear()
 
-        if rel == None:  # No relation to show
+        if rel is None:  # No relation to show
             self.ui.table.setColumnCount(1)
             self.ui.table.headerItem().setText(0, "Empty relation")
             return
@@ -246,7 +249,13 @@ class relForm(QtWidgets.QMainWindow):
         for i in rel.content:
             item = QtWidgets.QTreeWidgetItem()
             for j,k in enumerate(i):
-                item.setText(j, k)
+                if k is None:
+                    item.setBackground(j, QtGui.QBrush(QtCore.Qt.darkRed, QtCore.Qt.Dense4Pattern))
+                elif isinstance(k, (int, float)):
+                    item.setForeground(j, QtGui.QPalette().link())
+                elif not isinstance(k, str):
+                    item.setBackground(j, QtGui.QBrush(QtCore.Qt.darkGreen, QtCore.Qt.Dense4Pattern))
+                item.setText(j, str(k))
             self.ui.table.addTopLevelItem(item)
 
         # Sets columns
@@ -283,13 +292,13 @@ class relForm(QtWidgets.QMainWindow):
         filename = QtWidgets.QFileDialog.getSaveFileName(
             self, QtWidgets.QApplication.translate("Form", "Save Relation"),
             "",
-            QtWidgets.QApplication.translate("Form", "Relations (*.csv)")
+            QtWidgets.QApplication.translate("Form", "Json relations (*.json);;CSV relations (*.csv)")
         )[0]
         if (len(filename) == 0):  # Returns if no file was selected
             return
 
         relname = self.ui.lstRelations.selectedItems()[0].text()
-        self.user_interface.relations[relname].save(filename)
+        self.user_interface.store(filename, relname)
 
     def unloadRelation(self):
         for i in self.ui.lstRelations.selectedItems():
@@ -303,9 +312,15 @@ class relForm(QtWidgets.QMainWindow):
     def editRelation(self):
         from relational_gui import creator
         for i in self.ui.lstRelations.selectedItems():
-            result = creator.edit_relation(
-                self.user_interface.get_relation(i.text())
-            )
+            try:
+                result = creator.edit_relation(
+                    self.user_interface.get_relation(i.text())
+                )
+            except Exception as e:
+                QtWidgets.QMessageBox.warning(
+                    self, QtWidgets.QApplication.translate("Form", "Error"), str(e)
+                )
+                return
             if result != None:
                 self.user_interface.set_relation(i.text(), result)
         self.updateRelations()
@@ -344,7 +359,7 @@ class relForm(QtWidgets.QMainWindow):
         from relational_gui import creator
         result = creator.edit_relation()
 
-        if result == None:
+        if result is None:
             return
         name = self.promptRelationName()
 
@@ -380,7 +395,7 @@ class relForm(QtWidgets.QMainWindow):
             pass
 
     def showSurvey(self):
-        if self.Survey == None:
+        if self.Survey is None:
             self.Survey = surveyForm.surveyForm()
             ui = survey.Ui_Form()
             self.Survey.setUi(ui)
@@ -389,7 +404,7 @@ class relForm(QtWidgets.QMainWindow):
         self.Survey.show()
 
     def showAbout(self):
-        if self.About == None:
+        if self.About is None:
             self.About = QtWidgets.QDialog()
             ui = about.Ui_Dialog()
             ui.setupUi(self.About)
@@ -407,7 +422,7 @@ class relForm(QtWidgets.QMainWindow):
                 "",
                 QtWidgets.QApplication.translate(
                     "Form",
-                    "Relations (*.csv);;Text Files (*.txt);;All Files (*)"
+                    "Relations (*.json *.csv);;Text Files (*.txt);;All Files (*)"
                 )
             )
             filenames = f[0]
